@@ -25,6 +25,7 @@
 #define GLOBAL_METHOD_CACHE(c,m) (rb_bug("global method cache disabled improperly"), NULL)
 #endif
 
+static int vm_redefinition_check_flag(VALUE klass);
 static void rb_vm_check_redefinition_opt_method(const rb_method_entry_t *me, VALUE klass);
 
 #define object_id           idObject_id
@@ -82,9 +83,7 @@ rb_class_clear_method_cache(VALUE klass, VALUE arg)
 void
 rb_clear_cache(void)
 {
-    rb_warning("rb_clear_cache() is deprecated.");
-    INC_GLOBAL_METHOD_STATE();
-    INC_GLOBAL_CONSTANT_STATE();
+    ONLY_FOR_INTERNAL_USE("rb_clear_cache()");
 }
 
 void
@@ -468,6 +467,22 @@ rb_add_refined_method_entry(VALUE refined_class, ID mid)
     }
 }
 
+static void
+check_override_opt_method(VALUE klass, VALUE arg)
+{
+    ID mid = (ID)arg;
+    const rb_method_entry_t *me, *newme;
+
+    if (vm_redefinition_check_flag(klass)) {
+	me = lookup_method_table(RCLASS_ORIGIN(klass), mid);
+	if (me) {
+	    newme = rb_method_entry(klass, mid);
+	    if (newme != me) rb_vm_check_redefinition_opt_method(me, me->owner);
+	}
+    }
+    rb_class_foreach_subclass(klass, check_override_opt_method, (VALUE)mid);
+}
+
 /*
  * klass->method_table[mid] = method_entry(defined_class, visi, def)
  *
@@ -579,6 +594,11 @@ rb_method_entry_make(VALUE klass, ID mid, VALUE defined_class, rb_method_visibil
 
     VM_ASSERT(me->def != NULL);
 
+    /* check optimized method override by a prepended module */
+    if (RB_TYPE_P(klass, T_MODULE)) {
+	check_override_opt_method(klass, (VALUE)mid);
+    }
+
     return me;
 }
 
@@ -590,7 +610,7 @@ rb_method_entry_make(VALUE klass, ID mid, VALUE defined_class, rb_method_visibil
 	    recv_class = rb_ivar_get((klass), attached);	\
 	    hook_id = singleton_##hook;			\
 	}						\
-	rb_funcall2(recv_class, hook_id, 1, &arg);	\
+	rb_funcallv(recv_class, hook_id, 1, &arg);	\
     } while (0)
 
 static void
@@ -713,17 +733,19 @@ method_entry_get_without_cache(VALUE klass, ID id,
 	    ent->mid = id;
 
 	    if (UNDEFINED_METHOD_ENTRY_P(me)) {
-		ent->me = 0;
-		me = 0;
+		me = ent->me = NULL;
 	    }
 	    else {
 		ent->me = me;
 	    }
 	}
 	else if (UNDEFINED_METHOD_ENTRY_P(me)) {
-	    me = 0;
+	    me = NULL;
 	}
     }
+    else if (UNDEFINED_METHOD_ENTRY_P(me)) {
+	me = NULL;
+    }
 
     if (defined_class_ptr)
 	*defined_class_ptr = defined_class;
@@ -796,9 +818,9 @@ prepare_callable_method_entry(VALUE defined_class, ID id, const rb_method_entry_
     }
     else {
 	cme = (const rb_callable_method_entry_t *)me;
+	VM_ASSERT(callable_method_entry_p(cme));
     }
 
-    VM_ASSERT(callable_method_entry_p(cme));
     return cme;
 }
 
@@ -1014,7 +1036,7 @@ rb_export_method(VALUE klass, ID name, rb_method_visibility_t visi)
 
     if (UNDEFINED_METHOD_ENTRY_P(me) ||
 	UNDEFINED_REFINED_METHOD_P(me->def)) {
-	rb_print_undef(klass, name, 0);
+	rb_print_undef(klass, name, METHOD_VISI_UNDEF);
     }
 
     if (METHOD_ENTRY_VISI(me) != visi) {
@@ -1516,7 +1538,7 @@ rb_alias(VALUE klass, ID alias_name, ID original_name)
 	if ((!RB_TYPE_P(klass, T_MODULE)) ||
 	    (orig_me = search_method(rb_cObject, original_name, &defined_class),
 	     UNDEFINED_METHOD_ENTRY_P(orig_me))) {
-	    rb_print_undef(klass, original_name, 0);
+	    rb_print_undef(klass, original_name, METHOD_VISI_UNDEF);
 	}
     }
 
@@ -1530,10 +1552,10 @@ rb_alias(VALUE klass, ID alias_name, ID original_name)
     if (visi == METHOD_VISI_UNDEF) visi = METHOD_ENTRY_VISI(orig_me);
 
     if (orig_me->defined_class == 0) {
-	rb_method_entry_t *alias_me;
-
-	alias_me = rb_add_method(target_klass, alias_name, VM_METHOD_TYPE_ALIAS, (void *)rb_method_entry_clone(orig_me), visi);
-	alias_me->def->original_id = orig_me->called_id;
+	rb_method_entry_make(target_klass, alias_name, target_klass, visi,
+			     VM_METHOD_TYPE_ALIAS, NULL, orig_me->called_id,
+			     (void *)rb_method_entry_clone(orig_me));
+	method_added(target_klass, alias_name);
     }
     else {
 	rb_method_entry_t *alias_me;
@@ -1820,7 +1842,7 @@ rb_mod_modfunc(int argc, VALUE *argv, VALUE module)
 		me = search_method(rb_cObject, id, 0);
 	    }
 	    if (UNDEFINED_METHOD_ENTRY_P(me)) {
-		rb_print_undef(module, id, 0);
+		rb_print_undef(module, id, METHOD_VISI_UNDEF);
 	    }
 	    if (me->def->type != VM_METHOD_TYPE_ZSUPER) {
 		break; /* normal case: need not to follow 'super' link */
@@ -1837,7 +1859,9 @@ rb_mod_modfunc(int argc, VALUE *argv, VALUE module)
 int
 rb_method_basic_definition_p(VALUE klass, ID id)
 {
-    const rb_method_entry_t *me = rb_method_entry(klass, id);
+    const rb_method_entry_t *me;
+    if (!klass) return TRUE; /* hidden object cannot be overridden */
+    me = rb_method_entry(klass, id);
     return (me && METHOD_ENTRY_BASIC(me)) ? TRUE : FALSE;
 }
 
@@ -1847,9 +1871,9 @@ call_method_entry(rb_thread_t *th, VALUE defined_class, VALUE obj, ID id,
 {
     const rb_callable_method_entry_t *cme =
 	prepare_callable_method_entry(defined_class, id, me);
-    const rb_block_t *passed_block = th->passed_block;
+    VALUE passed_block_handler = vm_passed_block_handler(th);
     VALUE result = vm_call0(th, obj, id, argc, argv, cme);
-    th->passed_block = passed_block;
+    vm_passed_block_handler_set(th, passed_block_handler);
     return result;
 }
 
@@ -1894,7 +1918,7 @@ vm_respond_to(rb_thread_t *th, VALUE klass, VALUE obj, ID id, int priv)
     const rb_method_entry_t *const me =
 	method_entry_get(klass, resid, &defined_class);
 
-    if (!me) return TRUE;
+    if (!me) return -1;
     if (METHOD_ENTRY_BASIC(me)) {
 	return -1;
     }
@@ -1917,8 +1941,8 @@ vm_respond_to(rb_thread_t *th, VALUE klass, VALUE obj, ID id, int priv)
 	    }
 	    else if (!NIL_P(ruby_verbose)) {
 		VALUE location = rb_method_entry_location(me);
-		rb_warn("%"PRIsVALUE"%c""respond_to?(:%"PRIsVALUE") is"
-			" old fashion which takes only one parameter",
+		rb_warn("%"PRIsVALUE"%c""respond_to?(:%"PRIsVALUE") uses"
+			" the deprecated method signature, which takes one parameter",
 			(FL_TEST(klass, FL_SINGLETON) ? obj : klass),
 			(FL_TEST(klass, FL_SINGLETON) ? '.' : '#'),
 			QUOTE_ID(id));
